/*
 * file QueryCommand.java
 * 
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.QueryCommand
 *
 * (C) Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.stp.client.samples;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.ListIterator;

import javax.wvcm.PropertyNameList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.StpLocation.Namespace;
import com.ibm.rational.wvcm.stp.StpProvider.Domain;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqQuery;
import com.ibm.rational.wvcm.stp.cq.CqRecordType;
import com.ibm.rational.wvcm.stp.cq.CqResultSet;
import com.ibm.rational.wvcm.stp.cq.CqRowData;
import com.ibm.rational.wvcm.stp.cq.CqQuery.DisplayField;
import com.ibm.rational.wvcm.stp.cq.CqQuery.Filter.Operation;

/**
 * A sample CM API application for defining and executing ClearQuest queries
 * using a command-line interface. Command line arguments are
 * <dl>
 * <dt> <b>-name</b> <i>existing-query-location</i>
 * <dd>Specifies an existing query to execute and possibly modify and save.</dd>
 * <dt> <b>-save</b> [<i>new-query-location</i>]
 * <dd>Specifies that the query is to be saved after definition or modification
 * and before attempting execution. If no location is specified here it must be
 * specified by a <b>-name</b> clause</dd>
 * <dt> <b>-select</b> <i>display-field-specification</i>+
 * <dd>Specifies the (new) display fields for the query. Requires either
 * <b>-from</b> or <b>-name</b>
 * <dd>
 * <dt> <b>-from</b> <i>primary-record-type-location</i>
 * <dd>Specifies the primary record type for the query. Requires either
 * <b>-select</b> or <b>-name</b>. The <b>-save</b> location, if specified
 * with <b>-name</b>, must be different from the <b>-name</b> location for
 * this to be valid, since the primary type of an existing query cannot be
 * changed.</dd>
 * <dt> <b>-where</b> <i>filtering-expression-term</i>*
 * <dd>Specifies the (new) filtering expression for the query. A <b>-where</b>
 * clause with no filtering-expression-terms forces the query to be run without
 * a filtering expression.</dd>
 * <dt> <b>-min</b> <i>first-row-to-show</i>
 * <dd>The first row of the result set to be displayed; defaults to 1, the
 * first row. Must be positive.</dd>
 * <dt> <b>-max</b> <i>last-row-to-show</i>
 * <dd>the last row of the result set to be displayed; defaults to the last
 * row. If zero, the number of rows found by the query will be reported.
 * <dt> <b>-with</b> <i>filter-leaf-expression</i>+
 * <dd>Specifies a value (and, optionally an operation) to fill in the dynamic
 * filters of the query.</dd>
 * <dt>-count
 * <dd>Request a count of the number of rows matched by the query, even if not
 * all rows are returned (as specified by the <b>-max</b> parameter)</dd>
 * </dl>
 * 
 * </pre>
 * 
 * The implementation is divided into four methods: collectArgs, buildQuery,
 * executeQuery, and printResults. As a sample CM API application, the
 * buildQuery and executeQuery methods are the most instructive. To make the
 * example realistic, a syntax for specifying display fields and filters on the
 * command line was contrived. The parsing of this representation is implemented
 * in the QueryUtilities class.
 */
public class QueryCommand extends QueryUtilities
{
    /** The CqProvider instance used by the class and its instances */
    private static CqProvider g_provider = null;
    
    /** The target ClearQuest user database, derived from command-line input */
    private static String g_repo = "";

    /** -name operand: The name of an existing query */
    StpLocation m_oldName = null;
    /** -save operand: the name under which the query is to be saved */
    StpLocation m_saveName = null;
    /** -from operand: the CqRecordType defining the type of records to query */
    StpLocation m_primaryType = null;
    /** -select operand: Used to build the DisplayField[] for the query */
    List<String> m_select = null;
    /** -where operand: Used to build the filtering expression for the query */
    List<String> m_where = null;
    /** -with operand: Used to build the FilterLeaf[] for query parameters */
    List<String> m_with = null;
    /** -min operand: The first row of the result set to return */
    long m_min = 1;
    /** -max operand: The last row of the result set to return */
    long m_max = Long.MAX_VALUE;
    /** Whether or not to count all rows: set by presence of -count */
    CqQuery.ListOptions m_countRows = null;
    
    /**
     * Collects the command-line arguments into object fields. Variables for
     * unspecified arguments are left at their initial value. As the arguments
     * are collected, the resource names are converted to a complete CM API
     * StpLocation and from that the targeted ClearQuest database is determined.
     * 
     * @param args A String array containing the command-line arguments
     * @throws WvcmException if StpLocation objects cannot be constructed from
     *             the resource names on the command line.
     */
    void collectArgs(String[] args)
    throws WvcmException
    {
        ListIterator<String> argList = Arrays.asList(args).listIterator(0);
        
        while (argList.hasNext()) {
            String arg = argList.next();
            
            if (arg.equals("-url"))
                g_provider.setServerUrl(argList.next());
            else if (arg.equals("-name"))
                m_oldName = toSelector(Namespace.QUERY, argList.next());
            else if (arg.equals("-save")) {
                int next = argList.nextIndex();
                
                if (next < args.length && !args[next].startsWith("-")) {
                    m_saveName = toSelector(Namespace.QUERY, args[next]);
                    argList.next();
                } else
                    m_saveName = m_oldName;
            } else if (arg.equals("-from"))
                m_primaryType = toSelector(Namespace.RECORD, argList.next());
            else if (arg.equals("-min"))
                m_min = Long.parseLong(argList.next());
            else if (arg.equals("-max"))
                m_max = Long.parseLong(argList.next());
            else if (arg.equals("-select"))
                m_select = getArgList(argList);
            else if (arg.equals("-where"))
                m_where = getArgList(argList);
            else if (arg.equals("-with"))
                m_with = getArgList(argList);
            else if (arg.equals("-count"))
                m_countRows = CqQuery.COUNT_ROWS;
            else if (arg.equals("-user"))
                Utilities.g_user = argList.next();
            else if (arg.equals("-password"))
                Utilities.g_pass = argList.next();
        }
    }

    /**
     * Constructs a CqQuery proxy in which the query specification data
     * presented on the command line is stored. If -save was specified, the
     * proxy will use that location since the query will be executed from that
     * location. If -save was not specified but -name was, then that location is
     * used since either the results will come directly from the named query or
     * the query will be done anonymously; Otherwise, a dummy location is used.
     */
    CqQuery buildQuery()
    throws WvcmException
    {
        CqQuery query = null;

        StpLocation queryLoc = m_saveName != null? m_saveName: g_provider
            .userFriendlySelector(Domain.CLEAR_QUEST,
                                  Namespace.QUERY,
                                  "anonymous",
                                  g_repo);

        query = g_provider.cqQuery(queryLoc);
        
        // If an existing query was specified, read it's specification from
        // the database and use that data to initialize the query proxy.
        if (m_oldName != null) {
            CqQuery oldQuery = (CqQuery) g_provider.cqQuery(m_oldName)
                .doReadProperties(QUERY_PROPERTIES);
            
            if (m_saveName == null || m_saveName.equals(m_oldName)) {
                query = oldQuery;
            } else {
                query.setPrimaryRecordType(oldQuery.getPrimaryRecordType());
                query.setDisplayFields(oldQuery.getDisplayFields());
                query.setFiltering(oldQuery.getFiltering());
            }
        } else
            query.setFiltering(query.cqProvider().buildFilterNode(Operation.CONJUNCTION, 
                                                         new CqQuery.Filter[0]));
        
        // Now populate the query with modifications specified on the command
        // line.
        
        // If a primary record type was specified, read the required information
        // about it from the database so that field and filter specifications
        // can be converted properly.
        if (m_primaryType != null) {
            CqRecordType queryRecordType = 
                (CqRecordType) g_provider.cqRecordType(m_primaryType)
                    .doReadProperties(RECORD_TYPE_WITH_FIELDS);

            query.setPrimaryRecordType(queryRecordType);
        }

        // If a -select clause was specified, parse it into a DisplayField
        // array and add it to the query proxy.
        if (m_select != null) {
            DisplayField[] fields = new DisplayField[m_select.size()];
    
            for (int i=0; i < fields.length; ++i)
                fields[i] = parseDisplayField(query, m_select.get(i));
    
            query.setDisplayFields(fields);
        }
        
        // If a -where clause was specified, parse it into a Filter object
        // and add it to the query proxy.
        if (m_where != null) {
            query.setFiltering(parseFilter(query, m_where.iterator()));
        }
        
        return query;
    }

    /**
     * Executes the constructed query.
     * <p>
     * At this point, if the user followed the rules, we should be able to
     * execute the query. If -save was specified, we need to define the query in
     * the database and then use doExecute to get the results. If an existing
     * query is being executed unmodified, then we use doExecute as well;
     * Otherwise we use CqRecordType.doQuery() passing the data from the command
     * line to ClearQuest.
     */
    CqResultSet executeQuery(CqQuery input)
    throws WvcmException
    {
        CqQuery query = input;
        
        // Create the query if so directed by the input.
        if (m_saveName != null && !m_saveName.equals(m_oldName)) {
            System.out.println("Creating " + query + " as " + describe(query));
            query = query.doCreateQuery(QUERY_PROPERTIES);
            
            // Copy new name to caller's proxy
            input.setProperty(CqQuery.USER_FRIENDLY_LOCATION, 
                              query.getUserFriendlyLocation());
        }

        CqResultSet results;

        // Use doExecute if the query is completely defined in the database (or
        // will be once dirty properties in the proxy have been written). There
        // will be dirty properties only if the input has directed that 
        // modifications to an existing query are to be written back to the same
        // location.
        if (query.updatedPropertyNameList().getPropertyNames().length == 0
                                                       || m_saveName != null) {
            CqQuery.FilterLeaf[] params = null;
            
            // If a -with clause was specified, parse the given
            // parameter values into a FilterLeaf array for use in doExecute;
            // otherwise leave the dynamic parameters null.
            
            if (m_with != null) {
                params = new CqQuery.FilterLeaf[query.getDynamicFilters().length];
                String sep = "With parameter(s) ";
                
                for (int i=0; i < params.length; ++i) {
                    String param = i < m_with.size()? m_with.get(i): "";

                    params[i] = parseDynamicFilter(query, i, param);
                    
                    System.out.print(sep + params[i]);  sep = ", ";
                }
                
                System.out.println("...");
            }

            // Note: If the display field and/or filter has been modified, it 
            // will be written to the definition before the query executes as a
            // side-effect of either of these do-methods.
            results = query.doExecute(m_min, m_max-m_min+1, m_countRows, params);
        } else
            results = query.getPrimaryRecordType()
                        .doQuery(query.getDisplayFields(), 
                                 query.getFiltering(), 
                                 m_min, m_max-m_min+1, 
                                 m_countRows);

        return results;
    }
    
    /**
     * Display the rows returned by the query on the console
     * @param query The query that was executed.
     * @param results The CqResultSet containing the rows of the result
     * set.
     * @throws WvcmException If interactions with the server fail or have failed
     * to obtained the necessary information.
     */
    void printResults(CqQuery query, CqResultSet results)
    throws WvcmException
    {
        System.out.print("Query '" + describe(query) + "' found ");

        // Display the results
        if (m_countRows != null)
            System.out.print(results.getRowCount() + " rows ");
        
        System.out.println("...");
        
        if (results.hasNext()) {
            // Print the column headings
            DisplayField[] fields = query.getDisplayFields();
            System.out.print("Row");
            
            for (int i=0; i < fields.length; ++i)
                if (fields[i].getIsVisible())
                    System.out.print("\t" + fields[i].getLabel());
            
            System.out.println();
    
            // Print the rows
            while (results.hasNext()) {
                CqRowData data = (CqRowData)results.next();
                Object[] row = data.getValues();
                
                System.out.print(data.getRowNumber());
                
                for (int i=0; i < row.length; ++i)
                    System.out.print("\t" + row[i]);
                
                System.out.println();
            }
        }
    }
    
    /**
     * Assembles a String containing information about the query executed.
     * @param query A CqQuery proxy for the query to be described. Must define
     * the DISPLAY_FIELDS, PRIMARY_RECORD_TYPE, and FILTERING properties
     * if the query has not been saved; USER_FRIENDLY_LOCATION, otherwise.
     * @return A String containing the name of the query if saved or a stylized
     * SQL select statement describing the query if not.
     */
    private String describe(CqQuery query)
    throws WvcmException
    {
        PropertyNameList updated = query.updatedPropertyNameList();
        
        if (updated.getPropertyNames().length == 0)
            return query.getUserFriendlyLocation().toString();
        
        StringBuffer sb = new StringBuffer();
        DisplayField[] fields = query.getDisplayFields(); 

        sb.append("select");
        
        for (int i=0; i < fields.length; ++i)
            sb.append (" " + fields[i]);
        
        sb.append(" from ");
        sb.append(((CqRecordType)query.getPrimaryRecordType()).getDisplayName());
        
        String filter = query.getFiltering().toString();
        
        if (!filter.equals(""))
            sb.append(" where " + filter);
        
        return sb.toString();
    }
    
    void run(String[] args)
    throws WvcmException
    {
        collectArgs(args);
        CqQuery query = buildQuery();
        CqResultSet results = executeQuery(query);
        printResults(query, results);
    }

    /**
     * @param args
     * @throws Exception 
     */
    public static void main(String[] args) throws Exception
    {
        if (g_provider == null)
            g_provider = Utilities.getStaticProvider();
        
        g_repo = "";
        
        new QueryCommand().run(args);
        System.exit(0);
    }
    
    /**
     * Collects the arguments in the command line upto, but not including, the
     * first argument that begins with "-"
     * 
     * @param args The command line argument iterator positioned on the argument
     *            that precedes the arguments to be collected. Must not be
     *            <b>null</b>.
     * @return A List of the collected arguments. Will not be <b>null</b>, but
     *         may be empty.
     */
    private static List<String> getArgList(ListIterator<String> args)
    {
        List<String> list = new ArrayList<String>();
        
        while (args.hasNext()) {
            String arg = args.next();
            
            if (arg.startsWith("-")) {
                args.previous();
                break;
            }
            
            list.add(arg);
        }
        
        return list;
    }
    
    /**
     * Converts command-line input specifying a resource into an StpLocation,
     * adding domain and namespace if omitted by the user
     * 
     * @param namespace The expected namespace for the location. Must not be
     *            <b>null</b>.
     * @param str The location as input by the user. Expect simple name and
     *            possibly repository information. Must not be <b>null</b>.
     * @return An StpLocation specifying the same resource as the given string
     *         argument based on context.
     * @throws WvcmException
     */
    private static StpLocation toSelector(Namespace namespace, String str)
        throws WvcmException
    {
        StpLocation loc = g_provider.stpLocation(str);

        if (loc.getDomain() == Domain.NONE)
            loc = loc.recomposeWithDomain(Domain.CLEAR_QUEST);

        if (loc.getNamespace() == Namespace.NONE)
            loc = loc.recomposeWithNamespace(namespace);

        if (g_repo.equals(""))
            g_repo = loc.getRepo();
        else if (loc.getRepo().equals(""))
            loc = loc.recomposeWithRepo(g_repo);

        return loc;
    }
    
    private static final PropertyRequest QUERY_PROPERTIES = 
        new PropertyRequest(CqQuery.PRIMARY_RECORD_TYPE
                                                 .nest(RECORD_TYPE_WITH_FIELDS),
                            CqQuery.DISPLAY_FIELDS,
                            CqQuery.DYNAMIC_FILTERS,
                            CqQuery.FILTERING,
                            CqQuery.USER_FRIENDLY_LOCATION
        );
}
