/*
 * file ViewRecord.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.ViewRecord
 *
 * (C) Copyright IBM Corporation 2005, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.stp.client.samples;

import static com.ibm.rational.stp.client.samples.ExecuteQuery.setUserFriendlyLocation;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;
import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyRequestItem.NestedPropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName;
import com.ibm.rational.wvcm.stp.cq.CqAttachment;
import com.ibm.rational.wvcm.stp.cq.CqAttachmentFolder;
import com.ibm.rational.wvcm.stp.cq.CqFieldValue;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqRecord;
import com.ibm.rational.wvcm.stp.cq.CqFieldValue.ValueType;

/**
 * View the current state of a record
 */
public class ViewRecord {
    /**
     * An instance of ExecuteQuery.Viewer that allows (read-only) viewing of a
     * Record resource.
     */
    static class Viewer implements ExecuteQuery.Viewer {
        Viewer(CqProvider provider) {  m_provider = provider;  }
        
        /**
         * @see com.ibm.rational.stp.client.samples.ExecuteQuery.Viewer#view(com.ibm.rational.wvcm.stp.cq.CqRecord)
         */
        public JFrame view(CqRecord record)
        {
            if (record != null) try {
                record = (CqRecord)record.doReadProperties(RECORD_PROPERTIES);
                return showRecord("View: ", record, null);
            } catch (WvcmException ex){
                ex.printStackTrace();
            }
            
            return null;
        }
        
        /**
         * Displays the content of an Attachment resource in a text window.
         * 
         * @param attachment An Attachment proxy for the attachment to be
         *            displayed.
         */
        public void view(CqAttachment attachment)
        {
            if (attachment != null) try{
                File file = File.createTempFile("attach", "tmp");
                
                attachment.doReadContent(file.getAbsolutePath(), null);
                BrowserDataModel.showFile(attachment.getDisplayName(), file);
            } catch(Throwable ex) {
                Utilities.exception(null, "View Attachment", ex);
            }
        }
        
        /**
         * Displays the ALL_FIELD_VALUES property of a record resource in a
         * table. The columns of the table are determined by the content of the
         * {@link #fieldMetaProperties} array. The display of most objects is
         * implemented by the object's own toString() method.
         * 
         * @param title The title string for the window that contains the table
         * @param record The Record proxy for the record to be displayed. Must
         *            define the ALL_FIELD_VALUES property and the FieldValues
         *            in that property must define the meta-properties listed in
         *            the {@link #fieldMetaProperties} array.
         * @param future Additional window components to be displayed along with
         *            the property table. (Used by extensions to this example).
         * @return A RecordFrame structure containing the GUI components created
         *         by this method.
         * @throws WvcmException
         */
        RecordFrame showRecord(String title, 
                               CqRecord record, 
                               JComponent[] future) throws WvcmException 
        {
            final StpProperty.List<CqFieldValue<?>> fields = 
                record.getAllFieldValues();
            
            // Define a table model in which each row is a property of the
            // record resource and each column is a meta-property of the
            // property, such as its name, type, and value;
            TableModel dataModel = new AbstractTableModel() {
                private static final long serialVersionUID = 1L;
                public int getColumnCount() { return fieldMetaProperties.length; }
                public int getRowCount() { return fields.size();}
                public Object getValueAt(int row, int col) 
                    { 
                        try {
                            Object val = fields.get(row)
                                .getMetaProperty((MetaPropertyName<?>)
                                                 fieldMetaProperties[col].getRoot());
                            
                            if (val instanceof CqRecord)
                                return ((CqRecord)val).getUserFriendlyLocation()
                                    .getName();
                            else if (val instanceof CqAttachmentFolder)
                                return ((CqAttachmentFolder)val)
                                    .getAttachmentList().size()
                                    + " attachments";
                            else
                                return val;
                                
                        } catch(Throwable ex) {
                            if (ex instanceof StpException) {
                                return ((StpException)ex).getStpReasonCode();  
                              } else {
                                  String name = ex.getClass().getName();
                                  return name.substring(name.lastIndexOf(".")+1);
                              }
                        }
                    }
                public String getColumnName(int col)
                    { return fieldMetaProperties[col].getRoot().getName(); }
            };
            
            // Define the display layout
            final JTable table = new JTable(dataModel);
            final JPanel panel = new JPanel(new BorderLayout());
            final JPanel buttons = new JPanel(new FlowLayout());
            final JButton button = new JButton("View");
            final RecordFrame frame = 
                new RecordFrame(title + record.getUserFriendlyLocation().toString(),
                                table, fields);
    
            // Add a button for viewing a selected record or attachment field
            buttons.add(button, BorderLayout.SOUTH);
            button.setEnabled(false);
            button.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent arg0)
                    {
                        int[] selected = table.getSelectedRows();
                        
                        for (int i =0; i < selected.length; ++i) {
                            int row = selected[i];
                            if (isAttachmentList(fields, row)) {
                                view(selectAttachment(frame, fields, row, "View"));
                            } else {
                                view(getRecordReferencedAt(fields, row));
                            }
                        }
                    }
                });
            
            // Add more buttons (used by later examples) 
            if (future != null)
                for(int i = 0; i < future.length; ++i) buttons.add(future[i]);
    
            table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    
            // Ask to be notified of selection changes and enable the view button
            // only if a record-valued field or attachment field is selected
            ListSelectionModel rowSM = table.getSelectionModel();
            rowSM.addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (!e.getValueIsAdjusting()){
                        int[] selected = table.getSelectedRows();
                        button.setEnabled(false);
    
                        for (int i=0; i <selected.length; ++i)
                            if (getRecordReferencedAt(fields, selected[i]) != null
                                || isAttachmentList(fields, selected[i])) {
                                button.setEnabled(true);
                                break;
                            }
                    }
                }
            });
    
            panel.add(new JScrollPane(table), BorderLayout.CENTER);
            panel.add(buttons, BorderLayout.SOUTH);
            frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
            frame.setContentPane(panel);
            frame.setBounds(g_windowX += 10, g_windowY += 10, 600, 300);
            frame.setVisible(true);
            
            return frame;
        }
    
        protected CqProvider m_provider;
    }
    
    /** 
     * Properties to be requested from each record field valie, including
     * specific additional information for attachments 
     */
    static final PropertyRequest VALUE_PROPERTIES =
        new PropertyRequest(StpResource.USER_FRIENDLY_LOCATION,
                            CqAttachmentFolder.ATTACHMENT_LIST
                                .nest(CqAttachment.DISPLAY_NAME,
                                      CqAttachment.FILE_NAME,
                                      CqAttachment.FILE_SIZE,
                                      CqAttachment.DESCRIPTION));
    
    /** The field meta-properties to be requested and displayed */
    static final NestedPropertyName[] fieldMetaProperties = 
        new PropertyRequest(CqFieldValue.NAME, 
                            CqFieldValue.REQUIREDNESS, 
                            CqFieldValue.TYPE,
                            CqFieldValue.VALUE.nest(VALUE_PROPERTIES)).toArray();
                  
    /** 
     * The PropertyRequest to use when reading data from a record to be
     * displayed by this viewer. Note the level of indirection used to request
     * the meta-properties of the fields in the ALL_FIELD_VALUES list rather
     * than those meta-properties of the ALL_FIELD_VALUES property itself.
     */
    final static PropertyRequest RECORD_PROPERTIES =
        new PropertyRequest(CqRecord.USER_FRIENDLY_LOCATION,
                            CqRecord.ALL_FIELD_VALUES
                            .nest(StpProperty.VALUE.nest(fieldMetaProperties)));
    
    /**
     * Examines the property value of a field and, if it references a record,
     * returns a proxy for the referenced record. Otherwise it returns null.
     * @param fields The Property.List to examine.
     * @param row The index of the element in the list to examine.
     * @return A Record proxy if the field references a record; null otherwise
     */
    static CqRecord getRecordReferencedAt(StpProperty.List<CqFieldValue<?>> fields, 
                                          int row)
    {
        try {
            CqFieldValue field = fields.get(row);
            
            if (field.getFieldType() == ValueType.RESOURCE 
                && field.getValue() instanceof CqRecord)
                return (CqRecord)field.getValue();
        } catch (WvcmException ex) { ex.printStackTrace(); }
    
        return null;
    }
    
    /**
     * Whether or not the indicated field is an attachment field.
     * @param fields The Property.List to examine.
     * @param row The index of the element in the list to examine.
     * @return true iff the field at the given index is an attachment field
     */
    static boolean isAttachmentList(StpProperty.List<CqFieldValue<?>> fields, 
                                    int row)
    
    {
        if (row >= 0) try {
            CqFieldValue field = fields.get(row);
            
            return field.getFieldType() == ValueType.ATTACHMENT_LIST;
        } catch (WvcmException ex) { ex.printStackTrace(); }
    
        return false;
    }
    
    /**
     * Presents to the user a list of the attachments associated with a
     * specified field of a record and allows the user to select one.
     * 
     * @param frame The parent frame for the dialog generated by this method.
     * @param fields The Property.List to examine.
     * @param row The index of the element in the list to examine.
     * @param op A string identifying the operation that will be performed on
     *            the selected attachment.
     * @return An Attachment proxy for the selected attachment; null if the user
     *         chooses to make no selection.
     */
    static CqAttachment 
    selectAttachment(Component frame,
                     StpProperty.List<CqFieldValue<?>> fields,
                     int row,
                     String op)
    {
        CqFieldValue field = fields.get(row);
        
        try {
            CqAttachmentFolder folder = (CqAttachmentFolder)field.getValue();
            ResourceList<CqAttachment> attachments = setUserFriendlyLocation
                (folder.doReadProperties(ATTACHMENT_PROPERTIES)
                    .getProperty(CqAttachmentFolder.ATTACHMENT_LIST));
            
            if (attachments.size() > 0) {
                CqAttachment attachment =
                    (CqAttachment) JOptionPane
                        .showInputDialog(frame,
                                         "Choose an Attachment to " + op,
                                         op + " Attachment",
                                         JOptionPane.INFORMATION_MESSAGE,
                                         null,
                                         attachments.toArray(),
                                         attachments.get(0));
                
                return attachment;
            }
         } catch(Throwable t) { Utilities.exception(frame, op + " Attachment", t);}
    
        return null;
    }
    
    /**
     * The attachment properties to be displayed in the attachment selection
     * list generated by {@link #selectAttachment}.
     */
    final static PropertyRequest ATTACHMENT_PROPERTIES =
        new PropertyRequest(CqAttachmentFolder.ATTACHMENT_LIST
                                .nest(CqAttachment.DISPLAY_NAME,
                                      CqAttachment.FILE_NAME,
                                      CqAttachment.FILE_SIZE,
                                      CqAttachment.DESCRIPTION,
                                      CqAttachment.USER_FRIENDLY_LOCATION));
    
    /**
     * The main program for the ViewRecord example. Instantiates a Provider and
     * then invokes the ExecuteQuery example, passing in a version of Viewer 
     * that displays fields of a ClearQuest record.
     * @param args not used.
     */
    public static void main(String[] args)
    {
        try {
            CqProvider provider = Utilities.getProvider().cqProvider();
            ExecuteQuery.run("View Record", provider, new Viewer(provider));
        } catch(Throwable ex) {
            Utilities.exception(null, "View Record", ex);
            System.exit(0);
        }
    }
    
    /**
     * An extension of JFrame for the record field display, 
     * exposing to clients the JTable component of the frame and
     * the field list that is being displayed in the table.
     */
    static class RecordFrame extends JFrame
    {
        RecordFrame(String title,
                    JTable table,
                    StpProperty.List fields)
        {
            super(title);
    
            m_table = table;
            m_fields = fields;
        }
    
        JTable m_table;
        StpProperty.List<CqFieldValue<?>> m_fields;
        private static final long serialVersionUID = 1L;
    }
    
    /** X offset for the next window to be displayed */
    private static int g_windowX = 200;
    
    /** Y offset for the next window to be displayed */
    private static int g_windowY = 200;
}
